/*-
 * Copyright (c) 2011 Philip Schulz <phs@deadc0.de>
 *
 * Permission to use, copy, modify, and distribute this software for any purpose
 * with or without fee is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/sysctl.h>

#include <machine/bus.h>
#include <machine/resource.h>

#include <sys/rman.h>

#include <dev/iicbus/iicbus.h>
#include <dev/iicbus/iiconf.h>
#include "iicbus_if.h"

#include "glx.h"
#include "glx_iic.h"

/*******************************************************************************
 *
 * Prototypes and Declarations
 *
 ******************************************************************************/
/* Internal helper methods */
static uint8_t  glxiic_rd(struct glxiic_softc *sc, uint8_t offset);
static void     glxiic_wr(struct glxiic_softc *sc, uint8_t offset, uint8_t data);
static void     glxiic_handle_irq(struct glxiic_softc *sc);

static void     glxiic_rd_status(struct glxiic_softc *sc);

static void     glxiic_timeout(void *arg);
static void     glxiic_setup_timeout(struct glxiic_softc *sc);
static void     glxiic_poll(void *arg);
static void     glxiic_setup_polling(struct glxiic_softc *sc);

static void     glxiic_dumpregs(device_t dev);

/*******************************************************************************
 *
 * Sysctl Interface
 *
 ******************************************************************************/
static int debug_freq = 0;
static int debug_timeout = 0;

SYSCTL_NODE(_debug, OID_AUTO, glxiic, CTLFLAG_RW, 0, "Geode LX IIC Driver");
SYSCTL_INT(_debug_glxiic, OID_AUTO, freq, CTLFLAG_RW, &debug_freq, 0,
  "Bus frequency for debugging");
SYSCTL_INT(_debug_glxiic, OID_AUTO, timeout, CTLFLAG_RW, &debug_timeout, 0,
  "Interrupt timeout for debugging");

/*******************************************************************************
 *
 * Interrupt handler(s)
 *
 ******************************************************************************/
int
glxiic_intr_fast(void *arg) {
    glxiic_softc_t *sc = arg;

    glxiic_handle_irq(sc);

    /* See comment in sys/bus.h */
    return (FILTER_HANDLED);
}

void
glxiic_intr(void *arg) {
    glxiic_softc_t *sc = arg;

    GLXIIC_LOCK(sc);
    glxiic_handle_irq(sc);
    GLXIIC_UNLOCK(sc);
}

/*******************************************************************************
 *
 * Internal methods
 *
 ******************************************************************************/
static uint8_t
glxiic_rd(struct glxiic_softc *sc, uint8_t offset) {
    return bus_read_1(sc->io_res, offset);
}

static void
glxiic_wr(struct glxiic_softc *sc, uint8_t offset, uint8_t data) {
    bus_write_1(sc->io_res, offset, data);
}

static void
glxiic_rd_status(struct glxiic_softc *sc) {
    sc->smb_sts         = glxiic_rd(sc, SMB_STS);
    sc->smb_ctrl_sts    = glxiic_rd(sc, SMB_CTRL_STS);
    sc->smb_ctrl1       = glxiic_rd(sc, SMB_CTRL1);

#ifdef GLXIIC_DEBUG
    device_printf(sc->dev,
      "status = %#x, ctrl_status = %#x, ctrl1 = %#x, fsm = %#x\n",
      sc->smb_sts, sc->smb_ctrl_sts, sc->smb_ctrl1, sc->state);
#endif

    glxiic_wr(sc, SMB_CTRL_STS, BB);
    glxiic_wr(sc, SMB_STS, (SLVSTP | BER | NEGACK));
}

int
glxiic_setup_slave(device_t dev) {
    int rc = -1;

    device_printf(dev, "I2C Slave Mode not yet supported!\n");

    return (rc);
}

int
glxiic_reset_master(device_t dev) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    int rc = 0;
    uint8_t reg;
    uint16_t freq;

    /*
     * SMB_CTRL2[1] enables the SMB logic in the CS5536 and should probably be
     * set before all other registers as clearing the bit resets the I2C core.
     */
    glxiic_wr(sc, SMB_CTRL2, ~ SMBEN);
    glxiic_wr(sc, SMB_CTRL2, SMBEN);

    /* Disable I2C slave address matching. */
    glxiic_wr(sc, SMB_ADDR, 0x00);

    /*
     * Set up I2C frequency. According to the data sheet (pg. 400), the period
     * of the SCL signal (t_SCL) is given as:
     *
     * t_SCL = 2 * SCLFRQ * tCLK
     *
     * SMB_CTRL2[7:1] contains SCLFRQ[6:1]
     * SMB_CTRL3[7:0] contains SCLFRQ[14:7]
     */
    debug_freq ? (freq = debug_freq) : (freq = sc->freq);

    if (bootverbose)
        device_printf(dev, "Using Frequency of %i (%#x)\n", freq, freq);

    reg = ((freq << 1) & 0xfe);
    reg |= SMBEN; /* keep core enabled */
    glxiic_wr(sc, SMB_CTRL2, reg);

    reg = (freq >> 7) & 0xff;
    glxiic_wr(sc, SMB_CTRL3, reg);

    /* Enable SMB interrupts */
    reg = glxiic_rd(sc, SMB_CTRL1);
    reg = INTEN;
    glxiic_wr(sc, SMB_CTRL1, reg);

    /* Re-set internal state */
    sc->state       = idle;

    return (rc);
}

static void
glxiic_handle_irq(struct glxiic_softc *sc) {
    cv_signal(&sc->glxiic_cv);
}

#define DUMPREG(_dev, _reg) \
  device_printf(_dev, #_reg "\t%#x: %#x\n", _reg, glxiic_rd(DEVTOSOFTC(_dev), _reg))

static void
glxiic_dumpregs(device_t dev) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    uint8_t sts;

    GLXIIC_TRACEIN(dev);
    device_printf(dev, "state=%#x\n", sc->state);
    sts = glxiic_rd(sc, SMB_STS);
    if (sts & SDAST)
        DUMPREG(dev, SMB_SDA);
    device_printf(dev, "SMB_STS\t%#x: %#x\n", SMB_STS, sts);
    DUMPREG(dev, SMB_CTRL_STS);
    DUMPREG(dev, SMB_CTRL1);
    DUMPREG(dev, SMB_ADDR);
    DUMPREG(dev, SMB_CTRL2);
    DUMPREG(dev, SMB_CTRL3);
    GLXIIC_TRACEOUT(dev);
}

static void
glxiic_timeout(void *arg) {
    struct glxiic_softc *sc = (struct glxiic_softc *) arg;

    GLXIIC_DUMP_BOOTVERBOSE(sc->dev);

    /*
     * callout is set up in a way that guarantees that our glxiic_lock mutex has
     * been allocated before the callout routine, i.e. this function is called.
     * Hence, we don't need to lock/unlock our softc here.
     */
    sc->state = iic_timeout;

    if (bootverbose)
        device_printf(sc->dev, "Timeout!\n");

    cv_signal(&sc->glxiic_cv);
}

static void
glxiic_setup_timeout(struct glxiic_softc *sc) {
    struct timeval tv;
    int ticks;

    tv.tv_sec   = (debug_timeout ? debug_timeout : sc->timeout);
    tv.tv_usec  = 0;

    if (tv.tv_sec == 0)
        return;

    ticks       = tvtohz(&tv);

    callout_reset(&sc->glxiic_callout, ticks, &glxiic_timeout, sc);
}

static void
glxiic_poll(void *arg) {
    struct glxiic_softc *sc = (struct glxiic_softc *) arg;

    cv_signal(&sc->glxiic_cv);
}

static void
glxiic_setup_polling(struct glxiic_softc *sc) {
    struct timeval tv;
    int ticks, interval;

    interval = (debug_freq ? debug_freq : sc->freq) * 2;

    tv.tv_sec   = interval / 1000;
    tv.tv_usec  = (interval * 1000) - tv.tv_sec;

    if ((tv.tv_sec == 0) && (tv.tv_usec < (10 * 1000)))
        tv.tv_usec = 10 * 1000;

    ticks       = tvtohz(&tv);

    if (interval)
        callout_reset(&sc->glxiic_callout, ticks, &glxiic_poll, sc);
}

/*******************************************************************************
 *
 * I2C Bus Interface
 *
 ******************************************************************************/
int
glxiic_repeated_start(device_t dev, u_char slave, int timeout /* us */) {
    int rc;

    GLXIIC_TRACEIN1(dev, slave);

    rc = glxiic_start(dev, slave, timeout);

    GLXIIC_TRACEOUT1(dev, rc);

    return rc;
}

int
glxiic_start(device_t dev, u_char slave, int timeout /* us */) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    int rc = 0, running = 1;
    uint8_t reg;

    GLXIIC_LOCK(sc);
    GLXIIC_TRACEIN1(dev, slave);

    /*
     * Signalling a START condition works differently, depending on whether it's
     * a real START condition or a repeated START:
     * In case of a real START condition, the MASTER bit in SMB_STS is not set
     * and thus, we first need to request bus mastership.
     * If we're already the bus master, we can go ahead and signal the START
     * condition right away.
     *
     * Real START condition:
     * 1) Write (STASTRE | START) into SMB_CTRL1, then wait for IRQ.
     * 2) After IRQ, SMB_STS should have SDASST, STASTR and MASTER set. If so,
     *    write device address into SMB_SDA and wait for IRQ. If not, clear
     *    error flags.
     * 3) After IRQ, check if SDAST is set, then write device address and wait
     *    for IRQ. If SDAST is not set, some unknown error occurred.
     * 4) After IRQ, check if NEGACK in SMB_STS is set. If so, no device
     *    responded to the address and we should release the bus.
     *    If there was an acknowledge, at MASTER needs to be set. If XMIT is set
     *    as well, then enter write mode, else enter read mode.
     *
     * Repeated START condition: Same as above, except MASTER is already set and
     * thus, there is no interrupt between 2) and 3).
     */
    glxiic_setup_timeout(sc);

    do {
        // GLXIIC_DUMP_BOOTVERBOSE(sc->dev);

        glxiic_rd_status(sc);

        switch(sc->state) {
        case iic_read:
        case iic_write:
        case idle:
            reg = sc->smb_ctrl1;
            reg &= ~ (ACK | STOP);
            reg |= STASTRE | START;
            glxiic_wr(sc, SMB_CTRL1, reg);

            if ((sc->smb_sts & MASTER) && (sc->smb_sts & SDAST)) {
                glxiic_wr(sc, SMB_SDA, slave);

                sc->state = select_device;
            } else {
                sc->state = wait_master;
            }
            break;

        case wait_master:
            if ((sc->smb_sts & MASTER) && (sc->smb_sts & SDAST)) {
                glxiic_wr(sc, SMB_SDA, slave);

                sc->state = select_device;
            } else {
                sc->state = error;

                GLXIIC_DUMP(dev);

                running = 0;
            }
            break;

        case select_device:
            running = 0;

            if (sc->smb_sts & NEGACK) {
                if (bootverbose)
                    device_printf(dev, "No ACK for device %#x!\n", slave);

                reg = STASTR | NEGACK;
                glxiic_wr(sc, SMB_STS, reg);

                rc = EFAULT;
                sc->state = idle;
            } else if (!(slave & 0x01)
              && (sc->smb_sts & MASTER) && (sc->smb_sts & XMIT)
              && (sc->smb_sts & STASTR) && (sc->smb_sts & SDAST)) {
                sc->state = iic_write_stalled;
            } else if ((slave & 0x01)
              && (sc->smb_sts & MASTER) && !(sc->smb_sts & XMIT)
              && (sc->smb_sts & STASTR)) {
                sc->state = iic_read_stalled;
            } else {
                GLXIIC_DUMP(dev);

                sc->state = error;
            }
            break;

        case iic_timeout:
            rc = ETIMEDOUT;
            running = 0;
            break;

        default:
            device_printf(dev, "%s(): Invalid state %i!\n", __func__, sc->state);
            rc = ENXIO;
        }

        if (running)
            cv_wait(&sc->glxiic_cv, &sc->glxiic_lock);
    } while (running);

    if (sc->state != iic_timeout)
        callout_stop(&sc->glxiic_callout);

    GLXIIC_TRACEOUT1(dev, rc);
    GLXIIC_UNLOCK(sc);

    return (rc);
}

int
glxiic_stop(device_t dev) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    int rc = 0, retries = 5, running = 1;
    uint8_t data;

    GLXIIC_LOCK(sc);
    GLXIIC_TRACEIN(dev);

    /*
     * For some reason, the device doesn't seem to be raise an interrupt
     * when the STOP condition has been transferred on the bus. To work
     * around this, we're employing a slow polling mechanism.
     */
    do {
        glxiic_rd_status(sc);
        glxiic_setup_polling(sc);

        if (!(sc->smb_sts & XMIT) && (sc->smb_sts & SDAST)) {
            glxiic_wr(sc, SMB_CTRL1, sc->smb_ctrl1 | ACK);

            data = glxiic_rd(sc, SMB_SDA);
            if (bootverbose)
                device_printf(dev, "Had to read 0x%.02x before STOP\n", data);

            glxiic_wr(sc, SMB_CTRL1, sc->smb_ctrl1 & ~ ACK);
        }

        /*
         * I used to have this code in the read function. However, as it turnes out,
         * it's better to do this here: If we can read a last byte from the bus,
         * i.e. the SDAST bit is set, apparently the STOP condition is not generated
         * on the bus until that byte is read from the device.
         */
        if ((((sc->smb_sts & (MASTER | SDAST)) && !(sc->smb_sts & XMIT))
          || ((sc->smb_sts & (MASTER | XMIT)))) && !(sc->smb_ctrl1 & STOP)) {
            glxiic_wr(sc, SMB_CTRL1, sc->smb_ctrl1 | STOP);
        }

        if (sc->smb_sts & STASTR)
            glxiic_wr(sc, SMB_STS, STASTR);

        running = (retries-- > 0) &&
          ((sc->smb_sts & MASTER) || (sc->smb_ctrl_sts & BUSY) || (sc->smb_ctrl1 & STOP));

        if (running)
            cv_wait(&sc->glxiic_cv, &sc->glxiic_lock);

        if (sc->state == iic_timeout) {
            rc = ETIMEDOUT;
            break;
        }

        callout_stop(&sc->glxiic_callout);
    } while (running);

    sc->state = idle;
    if (retries <= 0) {
        rc = ETIMEDOUT;

        GLXIIC_DUMP_BOOTVERBOSE(sc->dev);

        if (bootverbose)
            device_printf(sc->dev, "Timed out waiting for STOP condition! "
              "Bus reset may be required.\n");

        sc->state = error;
    }

    GLXIIC_TRACEOUT(dev);
    GLXIIC_UNLOCK(sc);

    return (rc);
}

int
glxiic_write(device_t dev, const char *buf, int len, int *sent,
  int timeout /* us */) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    int rc = 0;
    *sent = 0;

    GLXIIC_LOCK(sc);
    GLXIIC_TRACEIN(dev);

    if ((sc->state != iic_write) && (sc->state != iic_write_stalled)) {
        rc = EBADF;
        goto out;
    }

    if (len < 1) {
        rc = EINVAL;
        goto out;
    }

    do {
        glxiic_setup_timeout(sc);

        glxiic_rd_status(sc);

        switch (sc->state) {
        case iic_write_stalled:
            if (!(sc->smb_sts & MASTER) || !(sc->smb_sts & XMIT)
              || !(sc->smb_sts & STASTR)) {
                if (bootverbose)
                    device_printf(dev, "%s(%i): Invalid H/W state %#x\n",
                      __func__, __LINE__, sc->smb_sts);

                sc->state = error;

                rc = ENXIO;
                goto out;
            }

            /* STASTRE (stall-after-start) is still set, clear it first */
            if (sc->smb_ctrl1 & STASTRE)
                glxiic_wr(sc, SMB_CTRL1, sc->smb_ctrl1 & (~ STASTRE));

            /* Unstall the bus by clearing STASTR */
            glxiic_wr(sc, SMB_STS, STASTR);

            sc->state = iic_write;
            /* FALLTHRU */

        case iic_write:
            if ((sc->smb_sts & NEGACK) || !(sc->smb_sts & MASTER)
              || !(sc->smb_sts & XMIT)) {
                /* Something went wrong on the bus, abort. */
                GLXIIC_DUMP(dev);

                glxiic_wr(sc, SMB_STS, (BER | NEGACK));
                glxiic_wr(sc, SMB_CTRL1, (ACK | STOP));

                sc->state = idle;
                goto out;
            } else if (!(sc->smb_ctrl_sts & TSDA)) {
#ifdef GLXIIC_DEBUG
                device_printf(dev, "Stalled\n");
#endif
            } else if (sc->smb_sts & SDAST) {
#ifdef GLXIIC_DEBUG
                device_printf(dev, "buf[%i] = 0x%.02x\n", *sent, ((int) buf[*sent]) & 0xFF);
#endif

                glxiic_wr(sc, SMB_SDA, buf[*sent]);
            }
            break;

        default:
            device_printf(dev, "%s(): Invalid state %i!\n", __func__, sc->state);
            rc = ENXIO;
            break;
        }

        cv_wait(&sc->glxiic_cv, &sc->glxiic_lock);
        if (sc->state == iic_timeout) {
            rc = ETIMEDOUT;
            break;
        }

        callout_stop(&sc->glxiic_callout);
    } while (++(*sent) < len);

out:
    if (sc->state != iic_timeout)
        callout_stop(&sc->glxiic_callout);

    GLXIIC_TRACEOUT1(dev, rc);
    GLXIIC_UNLOCK(sc);

    return (rc);
}

int
glxiic_read(device_t dev, char *buf, int len, int *read, int last,
  int delay /* us */) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    int rc = 0, running = 1;
    *read = 0;

    GLXIIC_LOCK(sc);
    GLXIIC_TRACEIN2(dev, len, last);

    if ((sc->state != iic_read) && (sc->state != iic_read_stalled)) {
        rc = EBADF;
        goto out;
    }

    if (len < 1) {
        rc = EINVAL;
        goto out;
    }

    do {
        glxiic_setup_timeout(sc);

        glxiic_rd_status(sc);

        switch (sc->state) {
        case iic_read_stalled:
            if (!(sc->smb_ctrl1 & STASTRE) || !(sc->smb_sts & STASTR)
              || !(sc->smb_sts & MASTER) || (sc->smb_sts & XMIT)) {
                if (bootverbose)
                    device_printf(dev, "%s(%i): Invalid H/W state: SMB_STS=%#x, "
                      "SMB_CTRL1=%#x\n", __func__, __LINE__,
                      sc->smb_sts, sc->smb_ctrl1);

                sc->state = error;

                rc = ENXIO;
                goto out;
            }

            /*
             * STASTRE (stall-after-start) is still set, clear it first.  Note:
             * If we're just reading a single byte, we need to set the ACK flag
             * right away in order to send a NACK after the next received byte.
             */
            sc->smb_ctrl1 &= ~ STASTRE;
            if (last && (len == 1)) {
                sc->smb_ctrl1 |= ACK;
            }
            glxiic_wr(sc, SMB_CTRL1, sc->smb_ctrl1);

            /* Unstall the bus */
            glxiic_wr(sc, SMB_STS, STASTR);

            sc->state = iic_read;
            break;
        case iic_read:
            if ((sc->smb_sts & BER) || (sc->smb_sts & NEGACK)) {
                /* Something went wrong on the bus, abort. */
                if (bootverbose)
                    device_printf(dev, "Bus error. SMB_STS=%#x\n", sc->smb_sts);

                GLXIIC_DUMP(dev);

                glxiic_wr(sc, SMB_STS, (BER | NEGACK));
                glxiic_wr(sc, SMB_CTRL1, (INTEN | ACK | STOP));

                sc->state = idle;

                rc = EIO;
                running = 0;
            }

            if (last) {
                if (((*read) + 2) >= len) {
                    /*
                     * If we're reading the second-to-last byte from the SDA
                     * register, we should set the ACK bit so the next receive,
                     * which is the last byte, is not acknowledged as demanded
                     * by the I2C standard.
                     *
                     * For the case where len == 1, we have set the ACK flag
                     * when we cleared out the STASTRE bit.
                     */
                    glxiic_wr(sc, SMB_CTRL1, sc->smb_ctrl1 | ACK);
                }
            }

            if (sc->smb_sts & SDAST) {
                buf[*read] = glxiic_rd(sc, SMB_SDA);

#ifdef GLXIIC_DEBUG
                device_printf(dev, "buf[%i, %i] = %#x\n", *read, last, (uint8_t) buf[*read]);
#endif

                running = (++(*read) < len);
            }
            break;
        default:
            device_printf(dev, "%s(): Invalid state %i!\n", __func__, sc->state);
            rc = ENXIO;
            break;
        }

        if (running)
            cv_wait(&sc->glxiic_cv, &sc->glxiic_lock);

        if (sc->state == iic_timeout) {
            rc = ETIMEDOUT;
            break;
        }

        callout_stop(&sc->glxiic_callout);
    } while (running);

out:
    GLXIIC_TRACEOUT(dev);
    GLXIIC_UNLOCK(sc);

    return (rc);
}

int
glxiic_reset(device_t dev, u_char speed, u_char addr, u_char *oldaddr) {
    struct glxiic_softc *sc = DEVTOSOFTC(dev);
    int rc = 0;
    uint8_t reg;

    GLXIIC_LOCK(sc);
    GLXIIC_TRACEIN(dev);

    rc = glxiic_reset_master(dev);

    glxiic_setup_timeout(sc);
    for (int i = 0; i < 8; i++) {
        reg = glxiic_rd(sc, SMB_CTRL_STS);
        if (!(reg & BUSY) && !(reg & BB) && (reg & TSDA))
            break;

        reg |= TGSCL;
        glxiic_wr(sc, SMB_CTRL_STS, reg);

        cv_wait(&sc->glxiic_cv, &sc->glxiic_lock);
        if (sc->state == iic_timeout) {
            sc->state = idle;
            break;
        }
    }
    if (sc->state != iic_timeout)
        callout_stop(&sc->glxiic_callout);

    GLXIIC_TRACEOUT(dev);
    GLXIIC_UNLOCK(sc);
    return (rc);
}

/*******************************************************************************
 *
 * Kernel Module Interface
 *
 ******************************************************************************/

/* Declare iicbus(4) that attaches to our driver */
DRIVER_MODULE(iicbus, glxiic, iicbus_driver, iicbus_devclass, NULL, NULL);

/* Module declaration */
MODULE_DEPEND(glxiic, iicbus, IICBUS_MINVER, IICBUS_MODVER, IICBUS_MAXVER);
MODULE_VERSION(glxiic, GLXIIC_VERSION);

