/*-
 * Copyright (c) 2011 Philip Schulz <phs@deadc0.de>
 *
 * Permission to use, copy, modify, and distribute this software for any purpose
 * with or without fee is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __GLX_IIC_H
#define __GLX_IIC_H

#include <sys/types.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/lock.h>
#include <sys/kernel.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/condvar.h>

/*******************************************************************************
 *
 * Prototypes
 *
 ******************************************************************************/
int     glxiic_start(device_t, u_char, int);
int     glxiic_repeated_start(device_t, u_char, int);
int     glxiic_stop(device_t);
int     glxiic_write(device_t, const char *, int, int *, int);
int     glxiic_read(device_t, char *, int, int *, int, int);
int     glxiic_reset(device_t, u_char, u_char, u_char *);

int     glxiic_intr_fast(void *arg);
void    glxiic_intr(void *arg);

int     glxiic_setup_slave(device_t dev);
int     glxiic_reset_master(device_t dev);

/*******************************************************************************
 *
 * Definitions
 *
 ******************************************************************************/
#define	GLXIIC_NAME	"glxiic"

/* Our own version */
#define GLXIIC_VERSION 1

typedef enum glxiic_state {
    idle                = 0,
    wait_master         = 1,
    select_device       = 2,
    iic_write_stalled   = 3,
    iic_write           = 4,
    iic_read_stalled    = 5,
    iic_read            = 6,
    error               = 7,
    iic_timeout         = 8
} glxiic_state_t;

typedef struct glxiic_softc {
    device_t            dev;

    /* Synchronization */
    struct mtx          glxiic_lock;
    struct cv           glxiic_cv;
    struct callout      glxiic_callout;

    /* State of driver's FSM */
    glxiic_state_t      state;

    /* Hardware State */
    uint8_t             smb_sts;
    uint8_t             smb_ctrl_sts;
    uint8_t             smb_ctrl1;

    /* I/O Port Resource */
    struct resource *   io_res;
    int                 io_rid;
    u_long              io_start;
    u_long              io_end;
    u_long              io_size;

    /* IRQ Resource */
    struct resource *   irq_res;
    int                 irq_rid;
    void *              irq_cookie;
    int                 irq_no; /* as specified in device.hints */

    /* Device Flags */
    int                 flags;
    int                 addr;
    int                 freq;
    int                 timeout;

    /* I2C Child Device */
    device_t            iicbus;
} glxiic_softc_t;

/* MSR Offsets */
#define MSR_GLD_CAPABILITIES        0x51400000
#define MSR_LBAR_SMB                0x5140000b
#define MSR_PIC_YSEL_LO             0x51400020
#define MSR_PIC_YSEL_HI             0x51400021
#define MSR_PIC_ZSEL_LO             0x51400022
#define MSR_PIC_ZSEL_HI             0x51400023

/*
 * SMB Controller Registers
 */
#define SMB_SDA         0x00    /* SMB Serial Data */

/* SMB Status */
#define SMB_STS         0x01
#define SLVSTP          (1 << 7)    /* Slave Stop (Read/Write 1 to Clear) */
#define SDAST           (1 << 6)    /* SMB_DATA Status (Read Only) */
#define BER             (1 << 5)    /* Bus Error */
#define NEGACK          (1 << 4)    /* Neg Acknowledge (Rd/Wr 1 to Clear) */
#define STASTR          (1 << 3)    /* Stall After Start */
#define NMATCH          (1 << 2)    /* New Match */
#define MASTER          (1 << 1)    /* Master (Read Only) */
#define XMIT            (1 << 0)    /* Transmit (Read Only) */

/* SMB Control Status */
#define SMB_CTRL_STS    0x02
#define TGSCL           (1 << 5)
#define TSDA            (1 << 4)
#define BB              (1 << 1)
#define BUSY            (1 << 0)

/* SMB Control 1 */
#define SMB_CTRL1       0x03
#define STASTRE         (1 << 7)    /* Stall After Start Enable. */
#define ACK             (1 << 4)
#define INTEN           (1 << 2)    /* Interrupt Enable */
#define STOP            (1 << 1)
#define START           (1 << 0)

/* SMB Address */
#define SMB_ADDR        0x04

/* SMB Control 2 */
#define SMB_CTRL2       0x05
#define SMBEN           (1 << 0)    /* Enable */

/* SMB Control 3 */
#define SMB_CTRL3       0x06

/* Reserved. */
#define SMB_RSVD        0x07

#define SCLFRQ_MIN      0x0008
#define SCLFRQ_MAX      0x7FFF

/* Flags that control driver behavior (set via device.hints) */
#define FLAG_SLAVE      0x0001  /* Set up as I2C Slave */
#define FLAG_400KHZ     0x0002  /* Set up 400 kHz */
#define FLAG_SOFTIRQ    0x0004  /* Don't route an IRQ, but poll instead */

/* Helper Macros */
#define GLXIIC_LOCK(_sc)    mtx_lock(&(_sc)->glxiic_lock)
#define GLXIIC_UNLOCK(_sc)  mtx_unlock(&(_sc)->glxiic_lock)

#define DEVTOSOFTC(dev) ((glxiic_softc_t *)device_get_softc(dev))

#define GLXIIC_DUMP_BOOTVERBOSE(_dev)       {                               \
    device_printf((_dev), "%s() @ %s:%i\n", __func__, __FILE__, __LINE__);  \
    glxiic_dumpregs((_dev));                                                \
}

#ifdef GLXIIC_DEBUG
#define GLXIIC_TRACEIN(_dev)    device_printf((_dev), "--> %s()\n", __func__)
#define GLXIIC_TRACEIN1(_d, _p) device_printf((_d), "--> %s(%#x)\n", __func__, _p)
#define GLXIIC_TRACEIN2(_d, _p, _q) device_printf((_d), "--> %s(%#x, %#x)\n", __func__, _q, _p)
#define GLXIIC_TRACEOUT(_dev)   device_printf((_dev), "<-- %s()\n", __func__)
#define GLXIIC_TRACEOUT1(_d, _r)    device_printf((_d), "<-- %s(rc=%#x)\n", __func__, _r)
#define GLXIIC_DUMP(_dev)       GLXIIC_DUMP_BOOTVERBOSE(_dev)
#else /* !GLXIIC_DEBUG */

#define GLXIIC_TRACEIN(_dev)
#define GLXIIC_TRACEIN1(_d, _p)
#define GLXIIC_TRACEIN2(_d, _p, _q)
#define GLXIIC_TRACEOUT(_dev)
#define GLXIIC_TRACEOUT1(_d, _r)
#define GLXIIC_DUMP(_dev)
#endif /* GLXIIC_DEBUG */

#endif /* !__GLX_IIC_H */

