/*-
 * Copyright (c) 2011 Philip Schulz <phs@deadc0.de>
 *
 * Permission to use, copy, modify, and distribute this software for any purpose
 * with or without fee is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/time.h>

#include <machine/bus.h>
#include <machine/resource.h>

#include <sys/rman.h>

#include <isa/isavar.h>

#include <dev/iicbus/iicbus.h>
#include <dev/iicbus/iiconf.h>
#include "iicbus_if.h"

#include "glx.h"
#include "glx_iic.h"

/*******************************************************************************
 *
 * Prototypes and Declarations
 *
 ******************************************************************************/
static void glxiic_isa_identify(driver_t *, device_t);
static int  glxiic_isa_probe(device_t);
static int  glxiic_isa_attach(device_t);
static int  glxiic_isa_detach(device_t);

static device_method_t glxiic_isa_methods[] = {
    /* device interface */
    DEVMETHOD(device_identify,          glxiic_isa_identify),
    DEVMETHOD(device_probe,             glxiic_isa_probe),
    DEVMETHOD(device_attach,            glxiic_isa_attach),
    DEVMETHOD(device_detach,            glxiic_isa_detach),

    /* iicbus interface */
    DEVMETHOD(iicbus_callback,          iicbus_null_callback),
    DEVMETHOD(iicbus_repeated_start,    glxiic_repeated_start),
    DEVMETHOD(iicbus_start,             glxiic_start),
    DEVMETHOD(iicbus_stop,              glxiic_stop),
    DEVMETHOD(iicbus_write,             glxiic_write),
    DEVMETHOD(iicbus_read,              glxiic_read),
    DEVMETHOD(iicbus_reset,             glxiic_reset),
    DEVMETHOD(iicbus_transfer,          iicbus_transfer_gen),

    { 0, 0 }
};

static devclass_t glxiic_isa_devclass;

static driver_t glxiic_isa_driver = {
    GLXIIC_NAME,
    glxiic_isa_methods,
    sizeof(glxiic_softc_t),
};

/* Internal helper methods */
static void     glxiic_isa_setup(device_t dev);
static int      glxiic_isa_read_hints(device_t dev);
static int      glxiic_isa_route_irq(device_t dev);

/*******************************************************************************
 *
 * ISA Bus Interface
 *
 ******************************************************************************/
static struct isa_pnp_id glxiic_ids[] = {
    {0, NULL}
};

static void
glxiic_isa_identify(driver_t *driver, device_t parent) {
    device_t    child;

    if (device_find_child(parent, GLXIIC_NAME, -1) != NULL)
        return;

    child = BUS_ADD_CHILD(parent, ISA_ORDER_SPECULATIVE, GLXIIC_NAME, 0);
    if (child == NULL) {
        device_printf(parent, "BUS_ADD_CHILD() failed for %s\n", GLXIIC_NAME);
    }

    glxiic_isa_setup(child);
}

static int
glxiic_isa_probe(device_t dev) {
    uint64_t            msr;
    glxiic_softc_t *    sc = DEVTOSOFTC(dev);
    int                 rc;
    device_t            parent = device_get_parent(dev);

    bzero(sc, sizeof(*sc));

    rc = ISA_PNP_PROBE(parent, dev, glxiic_ids);
    if (rc != ENOENT)
        return (ENXIO);

    msr = rdmsr(MSR_GLD_CAPABILITIES);
    switch (msr) {
    case 0x000000000005DF502LL:
        /* Ok, device is supported */
        glxiic_isa_setup(dev);
        break;
    default:
        if (bootverbose) {
            device_printf(dev, "Unsupported MSR=%#llx\n", msr);
        }

        return (ENXIO);
    }

    device_set_desc(dev, "AMD CS5536 SMBus/I2C controller");

    return (0);
}

static int
glxiic_isa_attach(device_t dev) {
    glxiic_softc_t *    sc = DEVTOSOFTC(dev);
    int                 rc = ENXIO;

    sc->dev = dev;

    mtx_init(&sc->glxiic_lock, device_get_nameunit(dev), GLXIIC_NAME,
      MTX_DEF | MTX_RECURSE);

    cv_init(&sc->glxiic_cv, "CS5536 hardware state change");

    callout_init_mtx(&sc->glxiic_callout, &sc->glxiic_lock, 0);

    rc = glxiic_isa_read_hints(dev);
    if (rc) {
        rc = ENXIO;
        goto out;
    }

    sc->io_res = bus_alloc_resource_any(dev, SYS_RES_IOPORT, &sc->io_rid,
      RF_ACTIVE);
    if (sc->io_res == NULL) {
        device_printf(dev, "bus_alloc_resource_any(9) failed!\n");
        goto out;
    }

    if (sc->flags & FLAG_SOFTIRQ)
        goto route_irq;

    sc->irq_rid = 0;
    sc->irq_res = bus_alloc_resource(dev, SYS_RES_IRQ, &sc->irq_rid,
      sc->irq_no, sc->irq_no, 1, RF_ACTIVE);
    if (sc->irq_res == NULL) {
        device_printf(dev, "Failed to allocate IRQ %i!\n", sc->irq_no);
        goto out;
    }

    /*
     * I don't think we can use the fast interrupt mechanism as it may open a
     * race condition between the driver routine writing H/W registers which are
     * causing interrupts and the interrupt handler itself: The driver routine
     * may not yet have reached the point of cv_wait() while the interrupt
     * handle already signalled the condition.
     */
#ifdef GLXIIC_FASTINTR
    rc = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_MISC | INTR_MPSAFE,
      glxiic_intr_fast, NULL, sc, &sc->irq_cookie);
    if (rc) {
#endif
        rc = bus_setup_intr(dev, sc->irq_res, INTR_TYPE_MISC | INTR_MPSAFE,
          NULL, glxiic_intr, sc, &sc->irq_cookie);
        if (rc) {
            device_printf(dev, "bus_setup_intr() failed!\n");
            goto out;
        }
#ifdef GLXIIC_FASTINTR
        device_printf(dev, "Failed to activate fast interrupt - "
          "using normal mode\n");
    }
#endif

route_irq:
    /* Route IRQ in the CS5536 Extended PIC after handlers have been set up. */
    rc = glxiic_isa_route_irq(dev);
    if (rc) {
        device_printf(dev, "Failed to route IRQ!\n");
        goto out;
    }

    if (sc->flags & FLAG_SLAVE) {
        rc = glxiic_setup_slave(dev);
        if (rc) {
            device_printf(dev, "Failed to setup slave mode!\n");
            goto out;
        }
    } else {
        sc->iicbus = device_add_child(dev, "iicbus", -1);
        if (sc->iicbus == NULL) {
            device_printf(dev, "device_add_child() failed!\n");
            goto out;
        }

        rc = glxiic_reset_master(dev);
        if (rc) {
            device_printf(dev, "Failed to setup master mode!\n");
            goto out;
        }
    }

    rc = bus_generic_probe(dev);
    if (rc) {
        device_printf(dev, "bus_generic_probe() failed with rc=%#x\n", rc);
        goto out;
    }

    rc = bus_generic_attach(dev);
    if (rc) {
        device_printf(dev, "bus_generic_attach() failed with rc=%#x\n", rc);
        goto out;
    }

    return (0);

out:
    rc = ENXIO;

    glxiic_isa_detach(dev);

    return (rc);
}

static int
glxiic_isa_detach(device_t dev) {
    glxiic_softc_t *sc = DEVTOSOFTC(dev);

    bus_generic_detach(dev);

    if (sc->iicbus) {
        device_delete_child(dev, sc->iicbus);

        sc->iicbus = NULL;
    }

    if (sc->io_res) {
        bus_release_resource(dev, SYS_RES_IOPORT, sc->io_rid, sc->io_res);

        sc->io_res = NULL;
        sc->io_rid = -1;
    }

    if (sc->irq_res) {
        if (sc->irq_cookie) {
            bus_teardown_intr(dev, sc->irq_res, sc->irq_cookie);
        }

        bus_release_resource(dev, SYS_RES_IRQ, sc->irq_rid, sc->irq_res);

        sc->irq_res = NULL;
        sc->irq_rid = 0;
    }

    cv_destroy(&sc->glxiic_cv);

    mtx_destroy(&sc->glxiic_lock);

    return (0);
}

/*******************************************************************************
 *
 * Internal methods
 *
 ******************************************************************************/
#define RES_ID      0x10
#define RES_SIZE    8

static void
glxiic_isa_setup(device_t dev) {
    glxiic_softc_t *sc = DEVTOSOFTC(dev);
    uint64_t        lbar;
    u_long          ioport, io_start, io_size;
    int             rc;

    lbar = rdmsr(MSR_LBAR_SMB);
    if (!(lbar & (1LL << 32))) {
        if (bootverbose) {
            device_printf(dev, "I/O memory mapping is not enabled by platform "
              "firmware!\n");
        }

        goto out;
    }

    ioport = (lbar & 0xFFF8);

    rc = bus_get_resource(dev, SYS_RES_IOPORT, RES_ID, &io_start, &io_size);

    if (sc != NULL) {
        sc->io_size     = RES_SIZE;
        sc->io_rid      = RES_ID;
        sc->io_start    = ioport;
        sc->io_end      = ioport + RES_SIZE;
    }

    if ((io_start == ioport) && (io_size == RES_SIZE)) {
        goto out;
    }

    rc = bus_set_resource(dev, SYS_RES_IOPORT, RES_ID, ioport, RES_SIZE);
    if (bootverbose && rc) {
        device_printf(dev, "bus_set_resource(9) for I/O Port failed with rc=%i\n", rc);
    }

out:
    return;
}

static int
glxiic_isa_read_hints(device_t dev) {
    glxiic_softc_t *sc = DEVTOSOFTC(dev);
    int rc, irq;

    rc = resource_int_value(GLXIIC_NAME, device_get_unit(dev), "flags",
      &sc->flags);
    if (rc) {
        if (bootverbose)
            device_printf(dev, "Failed to query flags (rc=%i), using defaults!\n", rc);
        sc->flags = 0;
    }

    if (bootverbose)
        device_printf(dev, "flags %#x\n", sc->flags);

    rc = resource_int_value(GLXIIC_NAME, device_get_unit(dev), "freq",
      &sc->freq);
    if (rc) {
        /*
         * If no frequency was set explicitly by the user, then fall back to
         * using default frequencies. They are based on the assumption that the
         * SMB/I2C Block is clocked by a 48 MHz clock. This is true for the
         * CS5536 chip on an Alix 1.C system (and probably for most other
         * designs based on the AMD reference design).
         */
        if (sc->flags & FLAG_400KHZ) {
            sc->freq = 0x20;
            if (bootverbose)
                device_printf(dev, "Using 400 kHz mode (freq=%#x)\n", sc->freq);
        } else {
            sc->freq = 0x71;
            if (bootverbose)
                device_printf(dev, "Using 100 kHz mode (freq=%#x)\n", sc->freq);
        }
    } else if (sc->freq < SCLFRQ_MIN) {
        device_printf(dev, "Frequency too high, using %#.8x!\n", SCLFRQ_MIN);
        sc->freq = SCLFRQ_MIN;
    } else if (sc->freq >= SCLFRQ_MAX) {
        device_printf(dev, "Frequency too low, using %#.8x!\n", SCLFRQ_MAX);
        sc->freq = SCLFRQ_MAX;
    }

    rc = resource_int_value(GLXIIC_NAME, device_get_unit(dev), "irq", &irq);
    if (rc) {
        device_printf(dev, "Failed to query IRQ number (rc=%i)!\n", rc);
        goto out;
    }

    if ((irq <= 0) || (irq > 15) || (irq == 2)) {
        device_printf(dev, "Invalid IRQ number %i!\n", irq);
        goto out;
    }

    sc->irq_no = irq;
    if (bootverbose)
        device_printf(dev, "Will route IRQ #%i\n", sc->irq_no);

    rc = resource_int_value(GLXIIC_NAME, device_get_unit(dev), "timeout",
      &sc->timeout);
    if (rc)
        sc->timeout = 0;

    if (sc->flags & FLAG_SLAVE) {
        rc = resource_int_value(GLXIIC_NAME, device_get_unit(dev), "addr",
          &sc->addr);
        if (rc) {
            device_printf(dev, "Failed to query I2C slave address (rc=%i)!\n", rc);
            goto out;
        }
    }

    rc = 0;

out:
    return (rc);
}

static int
glxiic_isa_route_irq(device_t dev) {
    glxiic_softc_t *sc = DEVTOSOFTC(dev);
    int             rc = 0, irq = sc->irq_no;
    uint32_t        msr;

    /*
     * According to Table 5-13 on pg. 111 of the CS5536 data sheet, the SMB
     * controller interrupt is on unrestricted source Y, input #12. The
     * soft-interrupt is unrestricted source Y, input #0.
     * Section 6.9.1.2 on pg. 381 contains information on how to decode the
     * respective IRQ mapper MSR.
     */
#define ROUTE_IRQ(_group, _irq, _msr)       \
    _msr &= ~ (0x0F << (4 * (_group % 8))); \
    _msr |= _irq << (4 * (_group % 8));

    msr = rdmsr(MSR_PIC_YSEL_HI);
    ROUTE_IRQ(12, irq, msr);
    wrmsr(MSR_PIC_YSEL_HI, msr);

    if (sc->flags & FLAG_SOFTIRQ) {
        msr = rdmsr(MSR_PIC_YSEL_LO);
        ROUTE_IRQ(0, irq, msr);
        wrmsr(MSR_PIC_YSEL_LO, msr);
    }

    return (rc);
}

/*******************************************************************************
 *
 * Kernel Module Interface
 *
 ******************************************************************************/

/* Declare ourselves */
DRIVER_MODULE(glxiic, isa, glxiic_isa_driver, glxiic_isa_devclass, NULL, NULL);

