/*-
 * Copyright (c) 2011 Philip Schulz <phs@deadc0.de>
 *
 * Permission to use, copy, modify, and distribute this software for any purpose
 * with or without fee is hereby granted, provided that the above copyright
 * notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/lock.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/mutex.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/malloc.h>

#include <machine/bus.h>
#include <machine/resource.h>

#include <sys/rman.h>

#include <dev/pci/pcivar.h>
#include <dev/pci/pcireg.h>

#include "glx.h"
#include "glx_isab.h"
#include "glx_iic.h"

/*******************************************************************************
 *
 * PCI-ISA Bridge Driver -- Prototypes and Declarations
 *
 ******************************************************************************/
static int                  glxisab_probe(device_t dev);
static int                  glxisab_attach(device_t dev);
static int                  glxisab_detach(device_t dev);

static struct resource *    glxisab_alloc_resource(device_t bus, device_t child,
                              int type, int *rid, u_long start, u_long end,
                              u_long count, u_int flags);

static int                  glxisab_release_resource(device_t dev,
                              device_t child, int type, int rid,
                              struct resource *r);

static device_method_t glxisab_methods[] = {
    /* device interface */
    DEVMETHOD(device_probe,             glxisab_probe),
    DEVMETHOD(device_attach,            glxisab_attach),
    DEVMETHOD(device_detach,            glxisab_detach),
    DEVMETHOD(device_shutdown,          bus_generic_shutdown),
    DEVMETHOD(device_suspend,           bus_generic_suspend),
    DEVMETHOD(device_resume,            bus_generic_resume),

    /* bus interface */
    DEVMETHOD(bus_print_child,          bus_generic_print_child),
    DEVMETHOD(bus_alloc_resource,       glxisab_alloc_resource),
    DEVMETHOD(bus_release_resource,     glxisab_release_resource),
    DEVMETHOD(bus_activate_resource,    bus_generic_activate_resource),
    DEVMETHOD(bus_deactivate_resource,  bus_generic_deactivate_resource),
    DEVMETHOD(bus_setup_intr,           bus_generic_setup_intr),
    DEVMETHOD(bus_teardown_intr,        bus_generic_teardown_intr),

    /* termination */
    { 0, 0 }
};

static driver_t glxisab_driver = {
    GLXISAB_NAME,
    glxisab_methods,
    sizeof(glxisab_softc_t)
};

devclass_t glxisab_devclass;

/*******************************************************************************
 *
 * Device Interface
 *
 ******************************************************************************/
static int
glxisab_probe(device_t dev) {
    if ((pci_get_class(dev) != PCIC_BRIDGE) ||
	(pci_get_subclass(dev) != PCIS_BRIDGE_ISA))
        goto out;

    switch(pci_get_devid(dev)) {
    case 0x20901022:
        break;
    default:
        goto out;
    }

    device_set_desc(dev, "PCI-ISA bridge (AMD CS5536 Southbridge)");
    return (0);

out:
    return (ENXIO);
}

static int
glxisab_attach(device_t dev) {
    glxisab_softc_t *sc = DEV2GLXISAB(dev);
    struct resource_list *rl;
    struct resource_list_entry *rle;
    int rc;

    resource_list_init(&sc->resources);

    rl = BUS_GET_RESOURCE_LIST(device_get_parent(dev), dev);
    if (rl == NULL) {
        device_printf(dev, "Failed to obtain resource list from parent!\n");
        goto out;
    }

    STAILQ_FOREACH(rle, rl, link) {
        resource_list_add(&sc->resources, rle->type, rle->rid, rle->start,
          rle->end, rle->count);
    }

    sc->isa = device_add_child(dev, "isa", 0);
    if (sc->isa == NULL) {
        device_printf(dev, "Failed to add child 'isa'!\n");
        goto out;
    }

    rc= bus_generic_attach(dev);

    return (rc);

out:
    glxisab_detach(dev);

    return (ENXIO);
}

static int
glxisab_detach(device_t dev) {
    glxisab_softc_t *sc = DEV2GLXISAB(dev);

    if (sc->isa) {
        device_delete_child(dev, sc->isa);
        sc->isa = NULL;
    }

    resource_list_free(&sc->resources);

    return (bus_generic_detach(dev));
}

/*******************************************************************************
 *
 * Bus Interface
 *
 ******************************************************************************/
static struct resource *
glxisab_alloc_resource(device_t bus, device_t child, int type, int *rid,
  u_long start, u_long end, u_long count, u_int flags) {
    struct glxisab_softc *sc = DEV2GLXISAB(bus);
    struct resource_list *rl = &sc->resources;
    struct resource_list_entry *rle;
    struct resource *res;

    /*
     * If a child requests one of the resource ranges that the CS5536 chip
     * announces via the BARs, then we need to request the resource in our name
     * from the parent PCI bus first.
     *  This is based on the assumption that this device is a child of the pci
     * bus and that the pci bus has allocated the I/O resources announced in the
     * BARs for us.
     *  All other requests for a resource are passed to the grand-parent so they
     * eventually end up at the root (nexus) -- this is what the original isab
     * driver does.
     */
    STAILQ_FOREACH(rle, rl, link) {
        if ((rle->type == type) && (rle->start == start)
          && (rle->end == end) && (rle->count == rle->count)) {
            if (rle->res != NULL)
                panic("glxisab_alloc_resource(): Resource already allocated!");

            goto found;
        }
    }

    return bus_generic_alloc_resource(bus, child, type, rid, start,
      end, count, flags);

found:
    res = bus_alloc_resource(bus, type, rid, start, end, count, flags);
    if (res) {
        rle->res = res;
        rman_set_device(res, child);
    }

    return res;
}

static int
glxisab_release_resource(device_t bus, device_t child, int type, int rid,
  struct resource *res) {
    struct glxisab_softc *sc = DEV2GLXISAB(bus);
    struct resource_list *rl = &sc->resources;
    struct resource_list_entry *rle;

    STAILQ_FOREACH(rle, rl, link) {
        if (rle->res == res) {
            goto found;
        }
    }

    return bus_generic_release_resource(bus, child, type, rid, res);

found:
    rle->res = NULL;

    rman_set_device(res, bus);

    return bus_release_resource(bus, type, rid, res);
}

/*******************************************************************************
 *
 * Kernel Module Interface
 *
 ******************************************************************************/
DRIVER_MODULE(glxisab, pci, glxisab_driver, glxisab_devclass, NULL, NULL);

